"""
Fetch the models at MODELS_URL and store them in MODELS_DIR, where the g2p-transcriber finds them.
"""

import os
import urllib.error
import zipfile
from io import BytesIO
from urllib.request import urlopen
from logging import getLogger

log = getLogger(__name__)
MODELS_URL = "https://github.com/grammatek/ice-g2p/releases/download/v1.1/ice-g2p-models.zip"
MODELS_DIR = 'fairseq_models'


def load_models():
    log.info(f'Downloading models from {MODELS_URL} ...')
    general_model_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), MODELS_DIR)
    try:
        with urlopen(MODELS_URL) as downl_file:
            zip_file = zipfile.ZipFile(BytesIO(downl_file.read()))
            log.debug("Extracting model ...")
            zip_file.extractall(general_model_path)
            log.debug("models extracted")
    except urllib.error.HTTPError:
        log.error(f'Could not download models from {MODELS_URL}')
        raise
    except urllib.error.URLError:
        log.error(f'Could not download models from {MODELS_URL}')
        raise


def main():
    load_models()


if __name__ == '__main__':
    main()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from ice_g2p import dictionaries
from ice_g2p import entry


VOWELS = ['a', 'á', 'e', 'é', 'i', 'í', 'o', 'ó', 'u', 'ú', 'y', 'ý', 'ö']
MODIFIER_MAP = dictionaries.get_modifier_map()
HEAD_MAP = dictionaries.get_head_map()
TRANSCR_MAP = dictionaries.get_dictionary()
MIN_COMP_LEN = 4
MIN_INDEX = 2       # the position from which to start searching for a head word


class CompoundTree:
    def __init__(self, pron_dict_entry):
        self.elem = pron_dict_entry
        self.left = None
        self.right = None

    def preorder(self):
        if not self.left:
            print(self.elem)
        if self.left:
            self.left.preorder()
        if self.right:
            self.right.preorder()


def contains_vowel(word):
    for c in word:
        if c in VOWELS:
            return True

    return False


def compare_transcripts(comp_transcript, head_transcript):
    """
    If a transcript differs only in a length mark or in voiced/voiceless or having post aspriation or not,
    it should be recognized as the same transcript (since we have already matched the corresponding word strings)
    """
    if not comp_transcript:
        return -1

    head_ind = len(head_transcript) - 1
    comp_ind = len(comp_transcript) - 1

    while head_ind >= 0:
        head_char = head_transcript[head_ind]
        comp_char = comp_transcript[comp_ind]
        if head_char == comp_char:
            head_ind -= 1
            comp_ind -= 1
        elif head_char == ':':
            head_ind -= 1
        elif comp_char == ':':
            comp_ind -= 1
        elif head_char == '0' or head_char == 'h':
            head_ind -= 2
        elif comp_char == '0' or comp_char == 'h':
            comp_ind -= 2
        else:
            return -1

    return comp_ind + 1 # make up for the last iteration where head_ind went below 0


def extract_transcription(entry, comp_head):
    """
    Get the transcript of comp_head from the pron. dictionary and try to match it with the transcript of
    the whole entry. We are somewhat flexible here: length symbol, voicelessness or postaspiration do not
    cause the matching to fail (a: == a, r_0 == r, t_h == t), since transcriptions might not always match 100%.

    Other variations could be added, like 'r t n' == 't n' like in barn: 'b a t n' vs. 'b a r t n', but we are
    not there yet.

    :param entry: PronDictEntry of the compound being analysed
    :param comp_head: the head of the compound as string
    :return:
    """

    head_transcr = TRANSCR_MAP[comp_head] if comp_head in TRANSCR_MAP else 'NO_TRANSCRIPT'
    head_syllable_index = entry.transcript.rfind(head_transcr)

    if head_syllable_index <= 0:
       head_syllable_index = compare_transcripts(entry.transcript, head_transcr)
    if head_syllable_index <= 0:
        #print("did not find transcription of " + comp_head + "!")
        #print("transcription in db: " + head_transcr + ", compound transcr: " + entry.transcript)
        return '', ''

    else:
        modifier_transcr = entry.transcript[0:head_syllable_index]
        head_transcr = entry.transcript[head_syllable_index:]
        return modifier_transcr, head_transcr


def lookup_compound_components(word):
    """
    Divides the word based on if its components are found in the compound database. The rule of thumb is that
    the longest possible head word shows the correct division, but if a modifier is found for a shorter
    head word, this one is chosen. If no modifier is found but a valid head word, the longest valid head word is
    returned, with the assumption that the modifier will also be valid, even if it is not in the dictionary.

    :param word:
    :return: extracted modifier and head if found, returns empty strings if no components are found
    """
    if len(word) <= MIN_COMP_LEN:
        return '', ''

    n = MIN_INDEX
    longest_valid_head = ''
    mod = ''
    while n < len(word) - 2:
        head = word[n:]
        if head in HEAD_MAP:
            if word[:n] in MODIFIER_MAP:
                return word[:n], head
            elif longest_valid_head == '':
                longest_valid_head = head
        n += 1

    if len(mod) == 0 and len(longest_valid_head) > 0:
        # assume we have a valid modifier anyway
        mod = word[:word.index(longest_valid_head)]
        # but only as long as it contains a vowel!
        if not contains_vowel(mod):
            mod = ''

    return mod, longest_valid_head


def extract_compound_components(comp_tree):
    """
    As long as compound components can be extracted, extract compound components and their transcripts recursively.

    :param comp_tree: a tree containing one root element. If compound components are found, they are added
    as children of the root.
    :return:
    """
    mod, head = lookup_compound_components(comp_tree.elem.word)
    if len(mod) > 0 and len(head) > 0:
        mod_transcr, head_transcr = extract_transcription(comp_tree.elem, head)
        if len(mod_transcr) > 0 and len(head_transcr) > 0:
            left_elem = entry.PronDictEntry(mod, mod_transcr)
            left_tree = CompoundTree(left_elem)
            comp_tree.left = left_tree
            right_elem = entry.PronDictEntry(head, head_transcr)
            right_tree = CompoundTree(right_elem)
            comp_tree.right = right_tree
            extract_compound_components(left_tree)
            extract_compound_components(right_tree)


def build_compound_tree(entry):
    """
    :param entry: a PronDictEntry
    :return: a binary tree based on compound division
    """

    comp_tree = CompoundTree(entry)
    extract_compound_components(comp_tree)
    return comp_tree




#!/usr/bin/env python3
# -*- coding: utf-8 -*-


class Syllable:
    """
    Syllabification processes phonetic transcripts of words, where each phone is separated by a space. This space
    separated transcription represents the content field of a Syllable object.
    Note that some phones might be written as two characters.
    """

    def __init__(self):
        self.content = '' #the transcription of the syllable, space between each phone
        self.has_nucleus = False
        self.cons_cluster = None
        self.stress = 0

    def __str__(self):
        return self.content

    def __repr__(self):
        return self.content

    def append(self, phone_str):
        self.content += phone_str + ' '

    def prepend(self, phone_str):
        self.content = phone_str + ' ' + self.content

    def last_phones(self, number=1):
        """
        return number last phones from content
        :param number: number of last phones to return
        :return:
        """
        phone_arr = self.content.split()
        if number <= len(phone_arr):
            result_phones = phone_arr[-number:]
            return ' '.join(result_phones)
        raise IndexError('Number of phones to large: ' + str(number)
                         + ' is larger than length of content (' + self.content + ')')

    def last_index(self):
        """
        :return: index of last character (not space) in content
        """
        cont_len = len(self.content.strip())
        ind = cont_len - 1 if cont_len > 0 else 0
        return ind

    def startswith(self, phone):
        if self.content[0] == phone:
            return True
        return False

    def endswith(self, phone):
        if self.content[self.last_index()] == phone:
            return True
        return False

    def index_of_cluster(self):
        if self.cons_cluster:
            return self.content.rfind(self.cons_cluster)
        else:
            return -1

    def remove_cluster(self):
        self.content = self.content[:self.index_of_cluster()]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
The pron_dict module processes an ASR pronunciation dictionary, containing entries of the form 'word     transcription',
computes the necessary additional information for TTS (part-of-speech, syllabification and stress marks) and
outputs the results in the CMU format or plain syllable format.

Example:
    Input entry:

    adolfsdóttir     a: t O l f s t ou h t I r

    CMU output entry:

    ("adolfsdóttir" n (((a:) 1) ((t O l f s) 3) ((t ou h) 1) ((t I r) 3))))

    Plain syllable output entry:

    adolfsdóttir - a:.t O l f s.t ou h.t I r


"""

__license__ = 'Apache 2.0 (see: LICENSE)'

from ice_g2p.syllabification import syllabify_tree_dict
from ice_g2p.tree_builder import build_compound_tree

from ice_g2p.entry import PronDictEntry


def init_pron_dict(dict_file):
    with open(dict_file) as f:
        dict_list = f.read().splitlines()

    tuples = [tuple(line.split('\t')) for line in dict_list]
    return init_pron_dict_from_tuples(tuples)


def init_pron_dict_from_tuples(tuples: list, syllab_symbol: str):
    pron_dict = {}
    for word, transcr in tuples:
        entry = PronDictEntry(word, transcr, syllab_symbol=syllab_symbol)
        pron_dict[word]=entry
    return pron_dict


def create_tree_dict(pron_dict):
    tree_dict = {}
    for word, entry in pron_dict.items():
        t = build_compound_tree(entry)
        tree_dict[word] = t
    return tree_dict


def syllabify_and_label(pron_dict):
    tree_dict = create_tree_dict(pron_dict)

    syllabified = syllabify_tree_dict(tree_dict)
    return syllabified


def syllabify_and_label_dict(dictfile):
    pron_dict = init_pron_dict(dictfile)
    return syllabify_and_label(pron_dict)


def syllabify_and_label_from_lists(word_list, transcr_list):
    pron_dict_pairs = list(zip(word_list, transcr_list))
    pron_dict = init_pron_dict_from_tuples(pron_dict_pairs)
    return syllabify_and_label(pron_dict)


def main():
    # add syllabification and stress labeling to a transcription
    # we need both the transcription and the original word string, the user has to make sure the indices
    # in the lists match, i.e. that a word at index n has its transcription at index n
    # returns a list of transcripts, syllabified only or syllabified and stress labeled
    #word_list = ['öskjuhlíð', 'yndislega', 'víetnam']
    #transcr_list = ['9 s c Y l_0 i D', 'I n t I s t l E G a','v i: j E t n a m']
    word_list = ['ferðast', 'sjö', 'komma', 'fimm', 'kílómetrar', 'í', 'tíu', 'gráður', 'hita']
    transcr_list = ['f E r D a s t', 's j 9:', 'k_h O m a', 'f I m', 'c_h i: l ou m E t r a r', 'i:', 't_h i: j Y',
                    'k r au: D Y r', 'h I: t a']
    syllab_with_stress = syllabify_and_label_from_lists(word_list, transcr_list)

    for entry in syllab_with_stress:
        print(entry.dot_format_syllables())

    for entry in syllab_with_stress:
        print(entry.stress_format())

    for entry in syllab_with_stress:
        print(entry.cmu_format())


    """
    # create a syllabified/stress labelled pronunciation dictionary from a plain transcribed dict
    dictfile = '../data/ice_pron_dict_standard_clear.csv'
    syllab_with_stress = syllabify_and_label_dict(dictfile)
    # print in cmu or in plain syllable format (syllable format not containin stress info):
    with open('../data/ice_pron_dict_standard_clear_cmu.txt', 'w') as f:
        for entry in syllab_with_stress:
            f.write(entry.cmu_format())
            f.write('\n')

    with open('../data/ice_pron_dict_standard_clear_syllab.txt', 'w') as f:
        for entry in syllab_with_stress:
            f.write(entry.syllable_format())
            f.write('\n')
    """


if __name__ == '__main__':
    main()



import os

package_path = os.path.dirname(os.path.abspath(__file__))
DICTIONARY_FILE = os.path.join(package_path, 'dictionaries/ice_pron_dict_standard_clear.csv')
HEAD_FILE = os.path.join(package_path, 'data/head_map.csv')
MODIFIER_FILE = os.path.join(package_path, 'data/modifier_map.csv')
VOWELS_FILE = os.path.join(package_path, 'data/vowels_sampa.txt')
CONS_CLUSTERS_FILE = os.path.join(package_path, 'data/cons_clusters_sampa.txt')


def read_map(filename):
    with open(filename) as f:
        file_content = f.read().splitlines()
    dict_map = {}
    for line in file_content:
        arr = line.split('\t')
        if len(arr) > 1:
            values = arr[1:]
        else:
            values = []
        key = arr[0]
        dict_map[key] = values
    return dict_map


def read_dictionary(filename):
    with open(filename) as f:
        file_content = f.read().splitlines()
    pronDict = {}
    for line in file_content:
        word, transcr = line.split('\t')
        pronDict[word] = transcr
    return pronDict


def read_list(filename):
    with open(filename) as f:
        file_content = f.read().splitlines()
    return file_content


def get_head_map():
    return read_map(HEAD_FILE)


def get_modifier_map():
    return read_map(MODIFIER_FILE)


def get_dictionary():
    return read_dictionary(DICTIONARY_FILE)


def get_vowels():
    return read_list(VOWELS_FILE)


def get_cons_clusters():
    return read_list(CONS_CLUSTERS_FILE)



#
# A complete grapheme-to-phoneme module for Icelandic.
# Transcribes graphemes (alphabetic text) to phonemes with or without the help of a pronunciation dictionary.
# Output is according to input: a wordlist as input produces a dictionary formatted output word\ttranscription,
# text as input produces a line by line transcription text\ttranscription.
#
# Syllabification and stress labeling is optional.
#
# Examples:
#
#

import os
import sys
import logging
import argparse
from pathlib import Path

from ice_g2p.converter import Converter
from ice_g2p.transcriber import Transcriber
from ice_g2p.transcriber import G2P_METHOD

AVAILABLE_DIALECTS = ['standard', 'north']


def write_transcribed(transcribed: dict, filename: Path, suffix: str, keep_original: bool) -> None:
    """
    Writes the transcriptions with the original grapheme strings to a file
    named 'filename' extended by 'suffix', for example:
    original filename: path/to/textfile.txt
    output filename: path/to/textfile_transcribed.tsv

    :param transcribed: a map with grapheme strings as keys and phonetic transcr. as values
    :param filename: the original filename containing the grapheme strings
    :param suffix: the suffix to label the output file with
    :param keep_original: write the original grapheme string in the first column
    :return:
    """
    filename_stem = filename.stem
    extended_filename = str(filename.parent) + '/' + filename_stem + suffix + '.tsv'
    with open(extended_filename, 'w') as f:
        for key in transcribed:
            if keep_original:
                f.write(key + '\t')
            f.write(transcribed[key] + '\n')


def process_string(input_str: str, dialect='standard', use_dict=False, syllab_symbol='', word_sep='',
                   stress_label=False, lang_detect=False) -> str:
    print('processing: "' + input_str + '"')
    g2p = Transcriber(G2P_METHOD.FAIRSEQ, dialect=dialect, lang_detect=lang_detect, syllab_symbol=syllab_symbol, word_sep=word_sep,
                      stress_label=stress_label, use_dict=use_dict)
    return g2p.transcribe(input_str)


def process_file(filename: Path, dialect='standard', use_dict=False, syllab_symbol='', word_sep='',
                   stress_label=False, lang_detect=False) -> dict:
    """
    Transcribes the content of 'filename' line by line
    :param filename: input file to transcribe
    :param word_sep: if the transcription should contain word separators
    :return: a map of grapheme strings and their phonetic transcriptions
    """
    print("processing: " + str(filename))
    with open(filename) as f:
        file_content = f.read().splitlines()

    g2p = Transcriber(G2P_METHOD.FAIRSEQ, dialect=dialect, use_dict=use_dict, syllab_symbol=syllab_symbol, word_sep=word_sep,
                   stress_label=stress_label, lang_detect=lang_detect)
    transcribed = {}
    for line in file_content:
        transcribed[line] = g2p.transcribe(line)

    return transcribed


def process_file_or_dir(file_or_dir: Path, out_suffix: str, dialect='standard', use_dict=False, syllab_symbol='', word_sep='',
                   stress_label=False, lang_detect=False, keep_original=False) -> None:
    print("processing: " + str(file_or_dir))
    if os.path.isdir(file_or_dir):
        for root, dirs, files in os.walk(file_or_dir):
            for filename in files:
                if filename.startswith('.'):
                    continue
                file_path = Path(os.path.join(root, filename))
                transcribed_content = process_file(file_path, dialect=dialect, use_dict=use_dict, syllab_symbol=syllab_symbol, word_sep=word_sep,
                   stress_label=stress_label, lang_detect=lang_detect)
                write_transcribed(transcribed_content, file_path, out_suffix, keep_original)
    elif os.path.isfile(file_or_dir):
        transcribed_content = process_file(file_or_dir, dialect=dialect, use_dict=use_dict, syllab_symbol=syllab_symbol, word_sep=word_sep,
                   stress_label=stress_label, lang_detect=lang_detect)
        write_transcribed(transcribed_content, file_or_dir, out_suffix, keep_original)


def convert(transcription: str, from_alpha: str, to_alpha):
    converter = Converter()
    converted = converter.convert(transcription, from_alpha, to_alpha)
    return converted


def get_alphabets():
    converter = Converter()
    return converter.get_valid_alphabets()


def get_arguments():
    parser = argparse.ArgumentParser(description='Transcribe text input to phonetic representation. Provide '
                                                 'an input file or directory, or a string on stdin to transcribe.')
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('--infile', '-if', type=Path, help='inputfile or directory')
    group.add_argument('--inputstr', '-i', help='input string')
    parser.add_argument('--dialect', '-a', default='standard',
                        help='dialect to transcribe by, available: "standard" and "north"')
    parser.add_argument('--sep', '-s', type=str, help='word separator to use')
    parser.add_argument('--syll', '-y', type=str, help='syllable separator to use')
    parser.add_argument('--stress', '-t', action='store_true', help='use stress labels')
    parser.add_argument('--dict', '-d', action='store_true', help='use pronunciation dictionary')
    parser.add_argument('--keep', '-k', action='store_true', help='keep original')
    parser.add_argument('--langdetect', '-l', action='store_true', help='use word-based language detection')
    parser.add_argument('--phoneticalpha', '-p', type=str, help='output in a specific phonetic alphabet')
    return parser.parse_args()


def main():
    args = get_arguments()
    keep_original = args.keep
    dialect = args.dialect
    word_sep = args.sep
    use_dict = args.dict
    syllab = args.syll
    stress = args.stress
    lang_detect = args.langdetect
    alphabet = args.phoneticalpha

    if dialect not in AVAILABLE_DIALECTS:
        logging.error(f'Transcription is not available for dialect "{dialect}". Available dialects: {AVAILABLE_DIALECTS}')
        sys.exit(1)

    available_alphabets = get_alphabets()
    if alphabet and not alphabet in available_alphabets:
        logging.error(f'{alphabet} is not available. Available phonetic alphabets: {available_alphabets}')
        sys.exit(1)

    # we need either an input file or directory, or a string from stdin
    if args.infile is not None:
        if not args.infile.exists():
            logging.error(str(args.infile) + ' does not exist.')
            sys.exit(1)
        else:
            process_file_or_dir(args.infile, '_transcribed', dialect=dialect, use_dict=use_dict, syllab_symbol=syllab, word_sep=word_sep,
                                stress_label=stress, lang_detect=lang_detect, keep_original=keep_original)

    if args.inputstr is not None:
        transcribed = process_string(args.inputstr, dialect=dialect, use_dict=use_dict, syllab_symbol=syllab, word_sep=word_sep,
                                stress_label=stress, lang_detect=lang_detect)

        if alphabet:
            transcribed = convert(transcribed, 'SAMPA', alphabet)
        if keep_original:
            print(args.inputstr + ' : ' + transcribed)
        else:
            print(transcribed)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from ice_g2p import entry

# max number of syllables for a word only having primary stress on first syllable, other syllables without stress
ONE_STRESS_SYLL_COUNT = 4

PRIMARY_STRESS = 1
NO_STRESS = 3

# grammatical and other endings containing a vowel and thus constituting a syllable of their own
ENDING_SYLLABLES = ['a', 'i', 'u', 'ar', 'ir', 'ur', 'is', 'um', 'na', 'ni', 'nu', 'ið', 'ins', 'sins', 'in', 'inn',
                    'unum', 'num', 'nna', 'nni', 'nnar', 'innar', 'nar', 'inum', 'nir', 'irnir', 'ina', 'va', 'var', 'vum',
                    'ngur', 'inu', 'stu', 'ra', 'ndi', 'da', 'di', 'un', 'uð', 'ri', 'gur', 'ga', 'gðar', 'gðir', 'gður',
                    'nga', 'leg', 'lega', 'nlegt', 'legt', 'semi', 'ning', 'arinnar', 'ði', 'tha', 'ðar', 'ðir', 'sti',
                    'nda', 'ba', 'ngu', 'inni', 'ður', 'ngum', 'ann', 'anna', 'anni', 'ara', 'ari', 'as', 'að', 'enn', 'í', 'ía']


def last_stress(pron_dict_entry):
    if len(pron_dict_entry.syllables) > 0:
        return pron_dict_entry.syllables[-1].stress
    else:
        return NO_STRESS


def same_word_with_ending(short_word, long_word):
    """

    :param short_word:
    :param long_word:
    :return: True if long_word is the same word as short_word, but with another grammatical ending, False otherwise
    """

    ending = long_word.word[len(short_word.word):].strip()
    if ending in ENDING_SYLLABLES:
        return True
    return False


def synchronize_stress(short_word, long_word):
    """
    Synchronizes the stress marks between the two words, whereas the longer word has to start
    with the shorter word. The stress marks of the shorter words are transferred to the beginning
    of the longer word.
    :param short_word: syllabified word with stress marks
    :param long_word: syllabified word with stress mark only on first syllable or none
    :return:
    """

    # dative of 'te' has two syllables, whereas longer words like 'teig' only have one syllable
    if short_word.word == 'tei':
        return

    for i, elem in enumerate(short_word.syllables):
        if i >= len(long_word.syllables) or i >= len(short_word.syllables):
            print('i: ' + str(i) + ' long: ' + str(long_word.syllables) + ' short: ' + str(short_word.syllables))
        else:
            long_word.syllables[i].stress = short_word.syllables[i].stress


def should_add_primary_stress(current_word, modifier):
    """
    Tests for plausibility of a primary stress on the first syllable after the common syllables of current word
    and its modifier. If the current word has more syllables, if the last syllable of the modifier does not
    have a primary stress and if current word is not simply the modifier with a grammatical ending, then this
    method returns True, False otherwise
    :param current_word: word starting with modifier
    :param modifier: the beginning of current word (possible compound modifier)
    :return: True if it is plausible that the syllable in current word that comes right after the common
    syllables of both words should have primary stress, False otherwise
    """
    if len(current_word.syllables) > len(modifier.syllables) and \
        last_stress(modifier) != PRIMARY_STRESS and not \
        same_word_with_ending(modifier, current_word):
            return True

    return False


def set_stress(syllabified_words):
    """
    Performs a simple stress-setting algorithm:
    - for every word with less than ONE_STRESS_SYLL_COUNT syllables, set primary stress on first syllable
    - for every longer word, check if it starts with a previous word, use the previous syllabification and
        add primary stress on the first syllable after the common syllables if it is not a flective ending:

        verslunar -> ve'rslunar
        verslunareigandi -> ve'rslunarei'gandi  (compound: primary stress on head word)
        verslunarinnar -> ve'rslunarinnar   (ending: no stress)

    :param syllabified_words: alphabetical sorted list of syllabified words
    :return: a list of syllabified words with stress marks on each syllable
    """
    result_list = []
    modifiers = []
    last_word = entry.PronDictEntry('')
    for current in syllabified_words:
        current.syllables[0].stress = PRIMARY_STRESS
        if current.word.startswith(last_word.word):
            synchronize_stress(last_word, current)
            if len(last_word.word) > 1:
                modifiers.append(last_word)
            if should_add_primary_stress(current, last_word):
                current.syllables[len(last_word.syllables)].stress = PRIMARY_STRESS
        else:
            mod_len = len(modifiers)
            while mod_len > 0:
                mod_len -= 1
                if current.word.startswith(modifiers[mod_len].word):
                    synchronize_stress(modifiers[mod_len], current)
                    if should_add_primary_stress(current, modifiers[mod_len]):
                        current.syllables[len(modifiers[mod_len].syllables)].stress = PRIMARY_STRESS
                    break
                else:
                    modifiers.pop()
                    mod_len = len(modifiers)

        result_list.append(current)
        last_word = current

    return result_list



"""
    Converts grapheme strings (texts) to phonetic transcriptions using a
    Fairseq transformer model.
"""

import os
from fairseq.models.transformer import TransformerModel

# if word separation is required in transcribed output
# use this separator
WORD_SEP = '-'
ALPHABET = '[aábcðdeéfghiíjklmnoóprstuúvxyýzþæö]'
ENGLISH_ALPHABET = '[aåäbcdefghijklmnoöpqrstuüvwxyz]'
DICT_PREFIX = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'dictionaries/ice_pron_dict_')


class FairseqG2P:

    def __init__(self, model_file='model-256-.3-s-s.pt', dialect='standard', use_english=False):
        """
        Initializes a Fairseq lstm g2p model according to model_path
        and model_file.
        :param model_file: the g2p model file
        :param dialect: the pronunciation variant to use
        :param use_cwd: if set to False, model_path has to be absolute
        """
        self.model_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'fairseq_models/ice-g2p-models', dialect)
        self.model_file = model_file
        if use_english:
            model_path_english = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'fairseq_models/ice-g2p-models', 'english')
            self.g2p_model = TransformerModel.from_pretrained(model_path_english, self.model_file)
            self.alphabet = ENGLISH_ALPHABET
        else:
            self.g2p_model = TransformerModel.from_pretrained(self.model_path, self.model_file)
            self.alphabet = ALPHABET
        self.pron_dict = self.read_prondict(dialect)
        self.custom_dict = None
        self.automatic_g2p_dict = {}

    def override_pron_dict(self, pron_dict: dict):
        """
        Override the core pronunciation dictionary initialized in init
        :param pron_dict: new pronunciation dictionary
        :return:
        """
        self.pron_dict = pron_dict

    def set_custom_dict(self, custom_dict: dict):
        """
        A custom dictionary will be used additionally to the built in dictionary.
        The custom dictionary, if present, is checked first and thus has priority over the built-in dictionary
        in the cases where the include the same words.
        Be careful when using a custom dictionary that it follows the same dialect as selected for the g2p and
        thus the built-in dictionary.

        :param custom_dict: a dictionary with custom vocabulary and/or transcriptions. Has priority over the
        built-in dicionary
        """
        self.custom_dict = custom_dict

    def transcribe(self, text, use_dict=False, sep=False) -> str:
        """
        Transcribes text according to the initialized transformer model.
        Text can be a single word or longer text.
        :param text: the text to transcribe
        :param sep: if True, inserts a separator between each transcribed word in text
        :return: transcribed version of text as string
        """
        transcribed_arr = []
        for wrd in text.split(' '):
            if not wrd:
                continue
            # start with lookup
            transcr = ''
            if use_dict:
                transcr = self.automatic_g2p_dict.get(wrd)
            if use_dict and self.custom_dict and not transcr:
                transcr = self.custom_dict.get(wrd)
            if use_dict and not transcr:
                transcr = self.pron_dict.get(wrd, '')
            if not transcr:

                # if transcription not yet found, perform automatic g2p
                if set(wrd).difference(self.alphabet):
                    print(text + ' contains non valid character(s) ' + str(
                        set(wrd).difference(self.alphabet)) + ', skipping transcription.')
                    continue

                transcr = self.g2p_model.translate(' '.join(wrd))

                # add to automatic_g2p_dict so that each word only gets transcribed once in batch processing.
                self.automatic_g2p_dict[wrd] = transcr

            # add transcription regardless of origin
            transcribed_arr.append(transcr)

        if sep:
            transcribed = WORD_SEP.join(transcribed_arr)
        else:
            transcribed = ' '.join(transcribed_arr)

        return transcribed

    @staticmethod
    def read_prondict(dialect: str) -> dict:
        dictfile = DICT_PREFIX + dialect + '_clear.csv'
        prondict = {}
        with open(dictfile) as f:
            content = f.read().splitlines()
        for line in content:
            wrd, transcr = line.split('\t')
            prondict[wrd] = transcr

        return prondict




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""

Perform syllabification on words transcribed with X-SAMPA.
Syllable rules for Icelandic as described in:
    - Anton Karl Ingason (2006): Íslensk atkvæði - vélræn nálgun. (http://www.linguist.is/skjol/atkvadi.pdf)
    - Kristján Árnason (2005): Hljóð. Ritröðin Íslensk tunga.

The basic assumption is that Icelandic syllable structure follows the onset-rhyme model. From this follows that,
whenever possible, a syllable has a 'need' for an onset consonant. This contradicts in some cases with rules for
'between-the-lines' word division rules saying that a word should be divided between lines
such that the next line starts with a vowel: 'hes.tur' vs. 'hest.ur'.

The following consonant combinations always build an onset together:

    s, p, t, k + v, j, r    (sv, sj, sr, pv, pj, pr, etc.)
    fr

# input: transcribed and aligned word, example: a v p r I G D I (afbrigði)
# output: syllabified word: af.prIG.DI (af.brig.ði)

# algorithm:
# 1) symbols upto the second vowel build the first syllable: 'afpr'
# 2) the remaining string is divided into syllables each starting with a vowel: 'afpr.IGD.I'
# 3) identify cons cluster: 'af(pr).IGD.I'
# 4) move consonants to onset (cons clusters and single consonants): 'af.prIG.DI'

This algorithm normally gives correct results for simple words, but can produce errors when applied to compounds.
Thus it is important to perform compound analysis before applying the core syllabification algorithm.

"""

from ice_g2p import syllable
from ice_g2p import dictionaries

# each syllable has a vowel as a nucleus. 'e' and 'o' aren't actually in the inventory, regardless of if we
# are using SAMPA or IPA, but we need to be able to identify 'ei' and 'ou' from the first character only, so those
# characters should be contained in the vowels list.
VOWELS = dictionaries.get_vowels()

# certain consonant clusters should not be split up between syllables
CONS_CLUSTERS = dictionaries.get_cons_clusters()

# Since we don't know which phonetic alphabet we are using, we check for either IPA or SAMPA representation of 'E'.
# If you are using a different alphabet, you need to add the definition of 'e' like in 'elda' of the respective
# alphabet, if it is different from SAMPA and IPA. Same goes for 'j' (which is identical for IPA and SAMPA).
VOWEL_E_SAMPA = 'E'
VOWEL_E_IPA = 'ɛ'
CONS_J = 'j'


def syllabify_on_nucleus(transcription_arr):
    """
    First round of syllabification. Divide the word such that each syllable
    starts with a vowel (except the first one, if the word starts with a consonant/consonants).
    """
    syllables = []
    current_syllable = syllable.Syllable()
    for phone in transcription_arr:
        if current_syllable.has_nucleus and phone in VOWELS:
            syllables.append(current_syllable)
            current_syllable = syllable.Syllable()

        if phone in VOWELS:
            current_syllable.has_nucleus = True

        current_syllable.append(phone)
    # append last syllable
    syllables.append(current_syllable)
    return syllables


def identify_clusters(entry):
    for syll in entry.syllables:
        for clust in CONS_CLUSTERS:
            if syll.content.strip().endswith(clust):
                syll.cons_cluster = clust


def syllabify_final(entry):
    """
    Iterate once more over the syllable structure and move consonants from rhyme to onset
    where appropriate, i.e. where one syllable ends with a consonant and the
    next one starts with a vowel. If a syllable ends with a consonant cluster, the whole
    cluster is moved to the next syllable, otherwise only the last consonant.
    However, if a syllable has a fixed boundary (as a result of decompounding), end or start,
    the boundary can not be changed.
    """

    for ind, syll in enumerate(entry.syllables):
        if ind == 0:
            continue
        prev_syll = entry.syllables[ind - 1]
        # syllable after the first syllable starts with a vowel - look for consonant onset in previous syllable
        # and move the consonant / consonant cluster from the previous to the current syllable
        if ind > 0 and syll.content[0] in VOWELS:
            if prev_syll.cons_cluster:
                # copy cons_cluster to next syllable
                syll.prepend(prev_syll.cons_cluster)
                prev_syll.remove_cluster()
                entry.update_syllables(ind, prev_syll, syll)
            elif prev_syll.last_phones() not in VOWELS:
                # handle 'jE' (=é) as one vowel
                if prev_syll.endswith(CONS_J) and (syll.startswith(VOWEL_E_SAMPA) or syll.startswith(VOWEL_E_IPA)):
                    phone = prev_syll.last_phones(1)
                    syll.prepend(phone)
                    prev_syll.content = prev_syll.content[:-(len(phone)+1)]
                else:
                    phone = prev_syll.last_phones()
                    syll.prepend(phone)
                    prev_syll.content = prev_syll.content[:-(len(phone)+1)]
                entry.update_syllables(ind, prev_syll, syll)


def syllabify_entry(entry):

    entry.syllables = syllabify_on_nucleus(entry.transcription_arr)
    identify_clusters(entry)
    syllabify_final(entry)


def syllabify_tree(entry_tree, syllables):
    """
    Recursively call syllabification on each element of entry_tree.
    Add up the syllables of the leaf nodes to build the syllable structure of the root element.
    :param entry_tree: a binary tree of a compound structure, the tree might not have any leaves,
    i.e. is not necessarily a compound
    :param syllables: an array to add up the leaf syllables of the tree
    :return:
    """
    if not entry_tree.left:
        syllabify_entry(entry_tree.elem)
        syllables += entry_tree.elem.syllables
    if entry_tree.left:
        syllabify_tree(entry_tree.left, syllables)
    if entry_tree.right:
        syllabify_tree(entry_tree.right, syllables)


def syllabify_tree_dict(tree_dict):
    """
    Syllabifies each entry in tree_dict
    :param tree_dict: a list of compoundTrees to syllabify
    :return: a list of syllabified PronDictEntries. Note that the returned list does NOT contain tree elements any more,
    but simple PronDictEntries.
    """
    syllabified = {}
    for word, t in tree_dict.items():
        syllables = []
        syllabify_tree(t, syllables)
        t.elem.syllables = syllables
        syllabified[word] = t.elem

    return syllabified






#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from ice_g2p.syllabification import VOWELS


class PronDictEntry:
    """
    Contains information on a pronunciation dict entry and methods to manipulate it

    The initialisation of a PronDictEntry object takes a word string and its transcription as parameters.
    """

    def __init__(self, word='', transcription='', reference='', alphabet='SAMPA', syllab_symbol='.'):
        """

        :param word: a dictionary entry (ex: 'dag')
        :param transcription: transcription of 'word', with the phones space separated (ex: 't a: G')

        """
        self.word = word
        self.transcript = transcription.strip()
        self.transcription_arr = self.transcript.split()
        self.transcript_variants = set()
        self.transcript_variants.add(self.transcript)
        self.gpos = 'nil'  # guessed part-of-speech
        # transcript, syllabification or stress labeled reference string for testing
        self.reference_transcr = reference
        self.syllables = []
        self.compound_elements = []
        self.entailing_compounds = [] # compounds where this word is one part
        self.frequency = 0
        self.set_phon_symbols(alphabet)
        self.syllab_symbol = syllab_symbol

    def __str__(self):
        return self.word + '\t' + self.gpos + '\t' + self.transcript + '\t' + str(self.syllables)

    def __repr__(self):
        return self.__str__()

    def set_phon_symbols(self, alphabet):
        if alphabet == 'IPA':
            self.length_symbol = 'ː'
            self.post_aspir = '\u02B0'
        else:
            self.length_symbol = ':'
            self.post_aspir = '_h'


    def simplify_compound_variants(self):
        tmp_variants = self.transcript_variants
        self.transcript_variants = set()
        orig_transcript = self.transcript
        no_postaspir = orig_transcript.replace(self.post_aspir, '')
        self.transcript_variants.add(orig_transcript)
        for t in tmp_variants:
            if t.replace(self.length_symbol, '') == orig_transcript or orig_transcript.replace(self.length_symbol, '') == t:
                continue
            elif t == no_postaspir:
                continue
            elif t.replace(self.length_symbol, '') == no_postaspir or no_postaspir.replace(self.length_symbol, '') == t:
                continue
            else:
                self.transcript_variants.add(t)

    def update_syllables(self, ind, prev_syll, syll):
        self.syllables[ind - 1] = prev_syll
        self.syllables[ind] = syll

    def cmu_format_syllables(self):
        formatted_string = '('
        for syll in self.syllables:
            formatted_string += '((' + syll.content + ') ' + str(syll.stress) + ') '

        return formatted_string.strip() + ')'

    def cmu_format(self):
        return '("' + self.word + '" ' + self.gpos + ' ' + self.cmu_format_syllables() + ')'

    def dot_format_syllables(self):
        sylls = ''
        for syll in self.syllables:
            sylls += syll.content.strip() + f" {self.syllab_symbol} "
        if len(sylls) > 3:
            # strip the last syllab_symbol with spaces from the word
            sylls = sylls[0:-3]
        return sylls

    def syllable_format(self):
        return self.word + ' - ' + self.dot_format_syllables()

    def stress_format(self):
        sylls = ''
        for syll in self.syllables:
            stressed_phones = []
            phones = syll.content.strip().split()
            for p in phones:
                if p in VOWELS:
                    p = p + str(syll.stress)
                stressed_phones.append(p)

            sylls += ' '.join(stressed_phones) + f" {self.syllab_symbol} "

        return sylls.strip()[:-2]

    def simple_stress_format(self):
        return self.stress_format()


"""
Converts phonetic transcriptions in one alphabet to transcriptions in another alphabet
"""
import os

package_path = os.path.dirname(os.path.abspath(__file__))
ALPHABETS_FILE = os.path.join(package_path, 'data/sampa_ipa_single_flite.csv')


class Converter:

    def __init__(self):
        self.alphabet_dictionary = self.init_dict()

    def init_dict(self):
        """
        Initialize a symbol dictionary where we can look up mappings between all alphabets in ALPHABETS_FILE.
        The result is a dictionary of dictionaries, where the top-level keys are the names of the alphabets
        and the sub-dictionaries the mappings from a top-level dictionary to all other dictionaries.

        Example:
        {'SAMPA' : {'a' : {'IPA' : 'a', 'SINGLE' : 'a', 'FLITE' : 'a'},
                   {'a:' : {'IPA' : 'aː', 'SINGLE' : 'A', 'FLITE' : 'aa'},
                   { ... }}
        'IPA'   : { ... },
        }

        """
        alphabet_dictionary = {}
        lines = self.read_mapping_file()
        headers = lines[0].split('\t')
        for ind, header in enumerate(headers):
            alphabet_dictionary[header] = {}
            for line in lines[1:]:
                symbols = line.split('\t')
                key_symbol = symbols[ind]
                symbol_dict = {}
                for i, h in enumerate(headers):
                    if i == ind:
                        continue
                    else:
                       symbol_dict[h] = symbols[i]
                alphabet_dictionary[header][key_symbol] = symbol_dict
        return alphabet_dictionary

    def read_mapping_file(self):
        with open(ALPHABETS_FILE) as f:
            lines = f.read().splitlines()
        return lines

    def get_valid_alphabets(self):
        return list(self.alphabet_dictionary.keys())

    def convert(self, input: str, from_alphabet: str, to_alphabet: str):
        """
        Converts a transcription in one alphabet to another alphabet.
        Raises a ValueError if either from or to alphabet is not available for conversion.
        If a symbol in the input string is not found in the respective from alphabet, the symbol is
        kept as is and not converted. A message is written to stdout if this happens.

        :param input: a transcribed string, the symbols have to be separated by a space character
        :param from_alphabet: the alphabet of the input string
        :param to_alphabet: the alphabet to convert the input string into
        :return: a converted transcription
        """

        valid = self.get_valid_alphabets()
        if from_alphabet not in valid or to_alphabet not in valid:
            raise ValueError(f"{from_alphabet} or {to_alphabet} is not contained in the converter's dictionary."
                             f" Valid alphabets: {valid}")

        if from_alphabet == to_alphabet:
            return input

        converted_str = ''
        current_dict = self.alphabet_dictionary[from_alphabet]
        for symbol in input.split():
            if symbol not in current_dict:
                print(f'"{symbol}" seems not to be a valid symbol in {from_alphabet} alphabet. Skipping conversion.')
                converted_str += symbol + ' '
            else:
                converted_symbol = current_dict[symbol][to_alphabet]
                converted_str += converted_symbol + ' '

        return converted_str.strip()


def main():
    converter = Converter()
    print(converter.get_valid_alphabets())
    print(converter.convert('p_h au f Y k l_0', 'SAMPA', 'IPA'))


if __name__ == '__main__':
    main()


"""G2P for Icelandic

Copyright (C) 2022 Grammatek ehf.

License: Apache 2.0 (see: LICENSE)
This package uses 3rd party submodules that might be published under other comparable open licenses.

This module is a grapheme-to-phoneme (g2p) program for Icelandic.
The _post_install() only gets called when installing from repository via setup.py install.
If installing from pip, it is necessary to run `fetch_models` before running g2p.

"""
import os
import urllib.error
import zipfile
from io import BytesIO
from urllib.request import urlopen
from logging import getLogger
from setuptools import setup

MODEL_URL = "https://github.com/grammatek/ice-g2p/releases/download/v1.1/ice-g2p-models.zip"

log = getLogger(__name__)


def _post_install():
    log.info(f'Downloading models from {MODEL_URL} ...')
    general_model_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'fairseq_models')
    try:
        with urlopen(MODEL_URL) as downl_file:
            zip_file = zipfile.ZipFile(BytesIO(downl_file.read()))
            log.debug("Extracting model ...")
            zip_file.extractall(general_model_path)
            log.debug("models extracted")
    except urllib.error.HTTPError:
        log.error(f'Could not download models from {MODEL_URL}')
        raise
    except urllib.error.URLError:
        log.error(f'Could not download models from {MODEL_URL}')
        raise


setup()

_post_install()


import math
from nltk import trigrams
from enum import Enum
import ice_g2p.syllab_stress_processing as syllabify
from ice_g2p.g2p_lstm import FairseqG2P
from ice_g2p.trigrams import ice_grams, eng_grams
from ice_g2p.stress import set_stress


class G2P_METHOD(Enum):
    FAIRSEQ = 1
    # THRAX = 2 TODO: implement!


class Transcriber:

    def __init__(self, g2p_method=G2P_METHOD.FAIRSEQ, dialect='standard', lang_detect=False, use_dict=False,
                 stress_label=False, syllab_symbol='', word_sep=''):

        self.g2p = self.init_g2p(g2p_method, dialect)
        self.use_dict = use_dict
        self.syllab_symbol = syllab_symbol
        self.word_separator = word_sep
        self.add_stress_label = stress_label
        if lang_detect:
            self.g2p_foreign = self.init_g2p(g2p_method, dialect=dialect, use_english=True)
            self.lang_detect = True
        else:
            self.g2p_foreign = None
            self.lang_detect = False
        if use_dict:
            from ice_g2p.dictionaries import get_dictionary
            self.dictionary = get_dictionary()
        else:
            self.dictionary = None

    def init_g2p(self, g2p_method: G2P_METHOD, dialect: str='standard', use_english=False) -> FairseqG2P:
        if g2p_method == G2P_METHOD.FAIRSEQ:
                return FairseqG2P(dialect=dialect, use_english=use_english)
        else:
            raise ValueError('Model ' + str(g2p_method) + ' does not exist!')

    def override_core_dict(self, pron_dict: dict):
        """
        Override the default pronunciation dictionary
        :param pron_dict:
        :return:
        """
        self.g2p.override_pron_dict(pron_dict)

    def set_custom_dict(self, custom_dict: dict):
        """
        A custom dictionary will be used additionally to the built in dictionary.
        The custom dictionary, if present, is checked first and thus has priority over the built-in dictionary
        in the cases where the include the same words.
        Be careful when using a custom dictionary that it follows the same dialect as selected for the g2p and
        thus the built-in dictionary.

        :param custom_dict: a dictionary with custom vocabulary and/or transcriptions. Has priority over the
        built-in dicionary
        """
        self.g2p.set_custom_dict(custom_dict)

    def transcribe(self, input_str: str, icelandic=True, cmu=False) -> str:
        transcr_arr = []
        for wrd in input_str.split(' '):
            if icelandic:
                # a word labelled as Icelandic will be sent to automatic lang detection
                transcr_arr.append(self.transcribe_lang(wrd.strip(), icelandic=self.is_icelandic(wrd.strip())))
            else:
                # word labelled as not Icelandic, will be sent directly to foreign transcription model
                transcr_arr.append(
                    self.transcribe_lang(wrd.strip(), icelandic=False))
        if self.syllab_symbol:
            entries = syllabify.init_pron_dict_from_tuples(list(zip(input_str.split(' '), transcr_arr)), self.syllab_symbol)
            syllabified_dict = syllabify.syllabify_and_label(entries)
            transcribed_utt = set_stress([syllabified_dict[wrd] for wrd in input_str.split(' ')])
            transcribed = self.extract_transcript(transcribed_utt, cmu)
        elif self.word_separator:
            transcribed = f' {self.word_separator} '.join(transcr_arr)
        else:
            transcribed = ' '.join(transcr_arr)

        return transcribed

    def extract_transcript(self, syllabified: list, cmu=False) -> str:
        if cmu:
            return self.extract_cmu_transcript(syllabified)
        result = ''
        for entr in syllabified:
            if self.add_stress_label:
                string_repr = entr.simple_stress_format()
            else:
                string_repr = entr.dot_format_syllables()
            if not result:
                result += string_repr
            elif self.word_separator:
                result += f" {self.word_separator} " + string_repr
            else:
                result += " " + string_repr

        return result

    def extract_cmu_transcript(self, syllabified: list) -> str:
        result = ''
        for entr in syllabified:
            if not result:
                result += entr.cmu_format()
            elif self.word_separator:
                result += f" {self.word_separator} " + entr.cmu_format()
            else:
                result += " " + entr.cmu_format()

        return result

    def transcribe_lang(self, input_str: str, icelandic=True) -> str:
        if icelandic or self.g2p_foreign is None:
            return self.g2p.transcribe(input_str.strip(), self.use_dict, self.word_separator)
        else:
            return self.g2p_foreign.transcribe(input_str.strip(), self.use_dict, self.word_separator)

    # Use trigrams to estimate the probability of a word being Icelandic or not
    def is_icelandic(self, word: str) -> bool:
        # if we don't have a foreign g2p, all words are processed as Icelandic
        if not self.lang_detect or not self.g2p_foreign:
            return True

        # If word contains non-valid characters for either of the models, it can't be transcribed by
        # the corresponding model. We use the Icelandic one as fallback, so just check for non-valid
        # English characters. Important check because of loanwords that might contain Icelandic characters
        # like: 'absúrd', 'dnépr', 'penélope' that have higher combined trigram probs for English despite
        # the non-valid trigrams containing Icelandic characters.
        if set(word).difference(self.g2p_foreign.alphabet):
            return True
        # Special case for spelling
        if len(word) == 1:
            return True
        ice_probs = []
        eng_probs = []
        for c1, c2, c3 in trigrams(word.lower(), pad_right=True, pad_left=True):
            if (c1, c2) in ice_grams and c3 in ice_grams[(c1, c2)]:
                ice_probs.append(ice_grams[(c1, c2)][c3])
            else:
                ice_probs.append(0.001)
            if (c1, c2) in eng_grams and c3 in eng_grams[(c1, c2)]:
                eng_probs.append(eng_grams[(c1, c2)][c3])
            else:
                eng_probs.append(0.001)
        if math.prod(ice_probs) >= math.prod(eng_probs):
            return True
        return False
